import ast
import error

from ast import Add, BinOp, Call, Constant, Expression, Mult, Name, Sub, expr
from collections.abc import Mapping
from dataclasses import dataclass

type Type = Int | Str | Fun


@dataclass
class Fun:
    params: list[Type]
    result: Type


@dataclass
class Int:
    ...


@dataclass
class Str:
    ...


def get_expr_type(e: expr, m: Mapping[str, Type]) -> Type:
    try:
        match e:
            case Name(id):
                if id in m:
                    return m[id]
                else:
                    raise NameError("name '" + id + "' is not defined")
            case Constant(i) if type(i) is int:
                return Int()
            case Constant(s) if type(s) is str:
                return Str()
            case BinOp(left, Add(), right):
                left_type: Type = get_expr_type(left, m)
                right_type: Type = get_expr_type(right, m)
                match left_type, right_type:
                    case Int(), Int():
                        return Int()
                    case Str(), Str():
                        return Str()
                    case _, _:
                        raise TypeError("unsupported operand type(s) for +")
            case BinOp(left, Sub(), right):
                left_type: Type = get_expr_type(left, m)
                right_type: Type = get_expr_type(right, m)
                match left_type, right_type:
                    case Int(), Int():
                        return Int()
                    case _, _:
                        raise TypeError("unsupported operand type(s) for -")
            case BinOp(left, Mult(), right):
                left_type: Type = get_expr_type(left, m)
                right_type: Type = get_expr_type(right, m)
                match left_type, right_type:
                    case Int(), Int():
                        return Int()
                    case Int(), Str():
                        return Str()
                    case Str(), Int():
                        return Str()
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")
            case Call(Name(f), args):
                match m.get(f):
                    case None:
                        raise NameError("name '" + f + "' is not defined")
                    case Fun(params, result):
                        types = [get_expr_type(arg, m) for arg in args]
                        if types == params:
                            return result
                        else:
                            raise TypeError("argument type mismatch")
                    case _:
                        raise TypeError("name '" + f + "' is not a function")
            case _:
                raise TypeError("Unsupported expression")
    except Exception as err:
        error.fail(e, err)


def get_type(e: Expression, m: Mapping[str, Type]) -> str:
    match get_expr_type(e.body, m):
        case Int():
            return "int"
        case Str():
            return "str"
        case Fun(_):
            return "<function>"


def f(x: str, y: int) -> str:
    return x * y



src = """
f(x, 2) * "e"
"""


def main() -> None:
    e: Expression = ast.parse(src, mode='eval')
    print(ast.dump(e, indent=4))
    print(src)
    print(get_type(e, {"x": Str(), "f": Fun([Str(), Int()], Int())}))


main()
